<?php

use AudentioForumConnect\HookInterface;
use AudentioForumConnect\Util\BbCodeParser;
use AudentioForumConnect\Util\Template;
use Composer\Autoload\ClassLoader;
use HaydenPierce\ClassFinder\ClassFinder;
use XFApi\Client;

/**
 * Class AudentioForumConnect
 */
class AudentioForumConnect
{
    /**
     * @var AudentioForumConnect
     */
    static $_instance;

    /**
     * @var
     */
    protected $pluginDir;

    public static $thread;
    public static $posts;
    public static $user;

    /**
     * @param $pluginDir
     * @return AudentioForumConnect
     * @throws Exception
     */
    public static function start($pluginDir)
    {
        self::$_instance = $instance = new self();
        $instance->pluginDir = $pluginDir;
        $instance->startAutoLoader();
        $instance->loadHooks();

        return $instance;
    }

    /**
     * @throws Exception
     */
    protected function loadHooks()
    {
        foreach (ClassFinder::getClassesInNamespace('AudentioForumConnect\Actions') as $action) {
            $action = new $action();

            if($action instanceof HookInterface) {
                add_action($action->getHook(), $action->getCallback(), $action->getPriority(), []);
            }
        }

        foreach (ClassFinder::getClassesInNamespace('AudentioForumConnect\Filters') as $filter) {
            $filter = new $filter();

            if($filter instanceof HookInterface) {
                add_filter($filter->getHook(), $filter->getCallback(), $filter->getPriority(), []);
            }
        }
    }

    /**
     * @param $phrase
     * @return string
     */
    public static function phrase($phrase) {
        return __($phrase, 'audentio-forum-connect');
    }

    /**
     * @return AudentioForumConnect
     */
    public static function getInstance()
    {
        return self::$_instance;
    }

    /**
     * @var
     */
    protected $autoLoader;

    /**
     *
     */
    protected function startAutoloader()
    {
        if ($this->autoLoader) {
            return;
        }

        /** @var ClassLoader $autoLoader */
        /** @noinspection PhpIncludeInspection */
        $autoLoader = require($this->pluginDir . '/vendor/autoload.php');
        $autoLoader->register();

        $this->autoLoader = $autoLoader;
    }

    /**
     * @var BbCodeParser
     */
    protected $bbCodeParser;

    /**
     * @return BbCodeParser
     */
    public function getBbCodeParser() {
        if(!$this->bbCodeParser) {
            $this->bbCodeParser = new BbCodeParser();
        }

        return $this->bbCodeParser;
    }

    /**
     * @return BbCodeParser
     */
    public static function bbCodeParser() {
        return self::getInstance()->getBbCodeParser();
    }

    /**
     * @var Client
     */
    protected $AudentioForumConnectClient;

    /**
     * @return Client
     */
    public function getAudentioForumConnectClient($option = null)
    {
        if (!$this->AudentioForumConnectClient) {
            if (!$option) {
                $option = get_option('wp_forum_connect');
            }
            $url = $option['xf_url'] ?? null;
            $key = $option['api_key'] ?? null;

            $this->AudentioForumConnectClient = new Client(trim($url, '/') . '/api', $key);
        }

        return $this->AudentioForumConnectClient;
    }

    /**
     * @return Client
     */
    public static function AudentioForumConnectClient($option = null)
    {
        return self::getInstance()->getAudentioForumConnectClient($option);
    }

    /**
     * @return bool
     */
    public static function isConfigured()
    {
        return (bool)self::AudentioForumConnectClient();
    }

    public static function checkConnection($outputMessage, $option = null)
    {
        if (!$option) {
            $option = get_option('wp_forum_connect');
        }

        if (empty($option['api_key'])) {
            return;
        }

        $client = self::AudentioForumConnectClient($option);

        try {
            if ($client->xf->thread->getThreads(1)) {
                update_option('wp_forum_connect_connection_error', null);
                if ($outputMessage) {
                    add_settings_error('xf_api_bridge', 'valid-credentials',
                        self::phrase('Successfully connected to your XenForo API.'), 'success');
                }
            }
        } catch (\Throwable $e) {
            $errorMessage = self::phrase('Failed to connect to your XenForo API. Check your credentials.');
            update_option('wp_forum_connect_connection_error', $errorMessage);
        }
    }

    public static function AudentioForumConnectClientForThread($userId)
    {
        $client = self::AudentioForumConnectClient();
        $option = get_option('wp_forum_connect');

        if ($apiKey = get_user_meta($userId, 'xf_api_key', true)) {
            $client->setApiKey($apiKey);

            return $client;
        }

        $xfUserId = get_user_meta($userId, 'xf_user_id', true);
        $client->setApiUserId($xfUserId);

        return $client;
    }

    public static function AudentioForumConnectClientForUser($userId = null, $returnGuest = false)
    {
        $client = self::AudentioForumConnectClient();
        $option = get_option('wp_forum_connect');

        if ($option['integration_type'] == 'wp_user') {
            $userId = $userId ?: get_current_user_id();
            if ($userId && $xfUserId = get_user_meta($userId, 'xf_user_id', true)) {
                $client->setApiUserId($xfUserId);

                return $client;
            }
        }

        if ($option['integration_type'] == 'session_cookie') {
            if (!empty($option['xf_cookie_prefix'])) {
                $client->setCookiePrefix($option['xf_cookie_prefix']);
            }
            if ($client->hasXfSession()) {
                if (!empty($option['session_auth_client_key'])) {
                    $client->setAuthenticationMethod('session_cookie');
                    $client->setApiKey($option['session_auth_client_key']);

                    return $client;
                } else {
                    $rememberCookie = $client->getUserCookie();
                    $sessionId = $client->getSessionCookie();

                    $xfUser = $client->xf->auth->fromSession($rememberCookie, $sessionId);
                    if ($xfUser) {
                        $client->setApiUserId($xfUser->user_id);
                        return $client;
                    }
                }
            }
        }

        if ($returnGuest) {
            return $client;
        }

        return null;
    }

    public static function verifyWpApiKey()
    {
        $headers = getallheaders();

        if (!isset($headers['X-Wpconnectapikey']) || $headers['X-Wpconnectapikey'] !== get_option('wp_forum_connect_wp_api_key')) {
            return false;
        }

        return true;
    }

    public static function use_xf_auth()
    {
        $option = get_option('wp_forum_connect');

        $commenting_mode = $option['commenting_mode'] ?? 'linked_threads';

        if ($commenting_mode == 'linked_threads') {
            return false;
        }

        return true;
    }

    /**
     * Functions for use in templates
    **/

    public static function get_xf_user()
    {
        if (self::$user) {
            return self::$user->getAttributes();
        }

        $client = self::AudentioForumConnectClientForUser(null, true);

        if ($client) {
            if (!$client->getApiUserId()) {
                $option = get_option('wp_forum_connect');

                if (!($option['default_forum_user_id'] ?? null)) {
                    return null;
                }

                $client->setApiUserId($option['default_forum_user_id']);
            }

            try {
                self::$user = $client->xf->me->me();

                return self::$user->getAttributes();
            } catch (\Exception $e) {
                return null;
            }
        }

        return null;
    }

    public static function xf_user_is_member_of($userGroupId)
    {
        $user = self::get_xf_user();

        if (!$user) {
            return false;
        }

        if (!$userGroupId)
        {
            return false;
        }

        if (is_string($userGroupId)) {
            $userGroupId = explode(',', $userGroupId);
        }

        if (is_array($userGroupId))
        {
            if (
                in_array($user['user_group_id'], $userGroupId)
                || array_intersect($userGroupId, $user['secondary_group_ids'])
            )
            {
                return true;
            }
        }
        else
        {
            if ($user['user_group_id'] == $userGroupId || in_array($userGroupId, $user['secondary_group_ids']))
            {
                return true;
            }
        }

        return false;
    }

    public static function should_create_thread($postId)
    {
        $option = get_option('wp_forum_connect');

        if ($option['commenting_mode'] == 'auth_only') {
            return false;
        }

        $createThread = false;
        $createThreadMetaData = get_post_meta($postId, 'xf_create_thread', true);
        if ($createThreadMetaData === '1' || ($createThreadMetaData === '' && !empty($option['create_thread']) && $option['create_thread'])) {
            $createThread = true;
        }

        return $createThread;
    }

    public static function has_xf_user()
    {
        $client = self::AudentioForumConnectClient();
        $option = get_option('wp_forum_connect');

        if ($option['integration_type'] == 'wp_user') {
            $userId = get_current_user_id();
            if ($userId && $xfUserId = get_user_meta($userId, 'xf_user_id', true)) {
                return true;
            }
        }

        if ($option['integration_type'] == 'session_cookie') {
            if (!empty($option['xf_cookie_prefix'])) {
                $client->setCookiePrefix($option['xf_cookie_prefix']);
            }
            if ($client->hasXfSession()) {
                return true;
            }
        }

        return false;
    }

    public static function get_xf_thread($id = null)
    {
        global $post;

        $client = self::AudentioForumConnectClient();

        $threadId = $id ?: self::get_post_thread_id($post->ID);

        return $client->xf->thread->getThread($threadId);
    }

    private static function load_thread_posts($postId)
    {
        $client = self::AudentioForumConnectClientForUser(null, true);

        if (!self::$thread) {
            self::$thread = $client->xf->thread->getThread(self::get_post_thread_id($postId));
        }
        if (!self::$posts) {
            self::$posts = $client->xf->thread->getThreadPosts(self::get_post_thread_id($postId));
        }
    }

    public static function get_xf_comments($postId)
    {
        self::load_thread_posts($postId);

        return self::format_xf_posts(self::$posts->posts);
    }

    public static function get_xf_featured_comments($postId)
    {
        if (!get_option('wp_forum_connect_plus')) {
            return;
        }

        $client = self::AudentioForumConnectClient();

        if (!self::$thread) {
            self::$thread = $client->xf->thread->getThread(self::get_post_thread_id($postId));
        }

        $posts = $client->audentio_wpconnect->thread->getThreadFeaturedPosts(self::get_post_thread_id($postId));

        return self::format_xf_posts($posts);
    }

    public static function get_xf_pagination($postId)
    {
        self::load_thread_posts($postId);

        return self::$posts->pagination;
    }

    public static function get_xf_comments_count($postId)
    {
        $threadPagination = self::get_xf_pagination($postId);
        return $threadPagination->total;
    }

    public static function get_post_thread_id($postId)
    {
        $option = get_option('wp_forum_connect');

        if (!empty($option['post_thread_id_meta_key'])) {
            if ($threadId = get_post_meta($postId, $option['post_thread_id_meta_key'], true)) {
                return $threadId;
            }
        }

        return get_post_meta($postId, 'xf_thread_id', true);
    }

    public static function update_post_thread_id($postId, $threadId)
    {
        update_post_meta($postId, 'xf_thread_id', $threadId);
    }

    public static function format_xf_posts($posts)
    {
        $return = [];
        $loggedIn = self::has_xf_user();

        foreach ($posts as $post) {

            if ($post->position == 0) {
                continue;
            }

            $formatted = [
                'id' => $post->post_id,
                'url' => $post->view_url,
                'message' => $post->message_parsed,
                'datetime' => $post->post_date,
                'datetime_iso_8601' => date('c', $post->post_date),
                'datetime_formatted' => date(get_option('date_format'), $post->post_date) . ' at ' . date(get_option('time_format'), $post->post_date),
                'wp_comment_id' => $post->wp_comment_id,
                'user' => [
                    'username' => $post->User['username'],
                    'url' => $post->User['view_url'],
                ],
                'can_reply' => $loggedIn,
                'reply_url' => self::$thread->view_url . 'reply' . (strpos(self::$thread->view_url, '?') === false ? '?' : '&') . 'quote=' . $post->post_id,
                'can_edit' => $post->can_edit,
                'edit_url' => $post->view_url . 'edit',
                'can_report' => $loggedIn,
                'report_url' => $post->view_url . 'report',
                'voting_enabled' => is_array($post->allowed_content_vote_types) && count($post->allowed_content_vote_types),
                'can_up_vote' => $post->can_content_vote && is_array($post->allowed_content_vote_types) && in_array('up', $post->allowed_content_vote_types),
                'can_down_vote' => $post->can_content_vote && is_array($post->allowed_content_vote_types) && in_array('down', $post->allowed_content_vote_types),
                'voted' => $post->visitor_content_vote,
                'vote_score' => $post->vote_score,
            ];

            $formatted['user']['avatar'] = self::get_user_avatar($post->User);

            $return[] = $formatted;
        }

        return $return;
    }

    public static function get_user_avatar($user)
    {
        $option = get_option('wp_forum_connect');

        $avatar = $avatar2x = plugins_url('article-forum-connect/public/img/avatar.png');

        if (!empty($user['email']) && ($option['email_gravatar'] ?? false)) {
            $avatar = 'https://secure.gravatar.com/avatar/' . md5($user['email']) . '?s=60&amp;d=mm&amp;r=g';
            $avatar2x = 'https://secure.gravatar.com/avatar/' . md5($user['email']) . '?s=120&amp;d=mm&amp;r=g';
        }

        if (!empty($user['gravatar'])) {
            $avatar = $avatar2x = $user['gravatar'];
        }
        if (!empty($user['avatar_urls']['o'])) {
            $avatar = $user['avatar_urls']['m'];
            $avatar2x = $user['avatar_urls']['l'];
        }

        return [
            'regular' => $avatar,
            'large' => $avatar2x,
        ];
    }

    public static function get_iframe_url($xfThread)
    {
        global $post;

        $option = get_option('wp_forum_connect');

        $data = [
            'url' => get_permalink($post),
            'open_comments' => $option['auto_open_iframe_query_string'] ?? '',
        ];

        $wpData = base64_encode(json_encode($data));

        $extraSegment = '';
        $postId = $_GET['post'] ?? null;
        $page = $_GET['comments-page'] ?? null;
        $unread = $_GET['unread'] ?? null;
        if ($postId) {
            $extraSegment = 'post-' . $postId;
        } elseif ($page) {
            $extraSegment = 'page-' . $page;
        //} elseif ($unread) {
        } else {
            $extraSegment = 'unread';
        }

        $theme = apply_filters('wp_forum_connect_iframe_theme', 'auto');

        $query = [];
        if (isset($_GET['add-reply'])) {
            $query[] = 'add-reply=1';
        }
        $query = '&' . implode('&', $query);

        return esc_url($xfThread->view_url . $extraSegment) . (strpos($xfThread->view_url, '?') === false ? '?' : '&') . 'in_iframe=1&theme=' . $theme . '&wp_data=' . $wpData . $query;
    }

    public static function auto_open_iframe($byDefault = false)
    {
        $option = get_option('wp_forum_connect');

        if (isset($option['auto_open_iframe']) && $option['auto_open_iframe']) {
            return true;
        }

        // we only want to check if it's open by default
        // if we're still here, it's only opened by query string, so not opened by default
        if ($byDefault) {
            return false;
        }

        if (isset($option['auto_open_iframe_query_string']) && $option['auto_open_iframe_query_string']) {
            if (strpos($option['auto_open_iframe_query_string'], '=') !== false) {
                list($key, $val) = explode('=', $option['auto_open_iframe_query_string']);

                if (isset($_GET[$key]) && $_GET[$key] == $val) {
                    return true;
                }
            }
        }

        return false;
    }

    public static function shortcode_xf_user_state()
    {
        if ($xfUser = self::get_xf_user()) {
            $xfUser['avatar'] = self::get_user_avatar($xfUser);

            echo Template::render('xf_user_state_logged_in', ['xfUser' => $xfUser]);
        } else {

            $option = get_option('wp_forum_connect');
            $loginUrl = $option['xf_url'] . '/login';
            $registerUrl = $option['xf_url'] . '/register';

            echo Template::render('xf_user_state_logged_out', ['loginUrl' => $loginUrl, 'registerUrl' => $registerUrl]);
        }
    }
}